#include <ftl/depth_camera.hpp>
#include "depth_camera_cuda.hpp"
#include <opencv2/core/cuda_stream_accessor.hpp>

using ftl::voxhash::DepthCamera;
using ftl::voxhash::DepthCameraCUDA;

DepthCamera::DepthCamera() {
	depth_mat_ = nullptr;
	colour_mat_ = nullptr;
	point_mat_ = nullptr;
	normal_mat_ = nullptr;
	depth_tex_ = nullptr;
	colour_tex_ = nullptr;
	normal_tex_ = nullptr;
}

void DepthCamera::alloc(const DepthCameraParams& params, bool withNormals) { //! todo resizing???
	depth_mat_ = new cv::cuda::GpuMat(params.m_imageHeight, params.m_imageWidth, CV_32FC1);
	colour_mat_ = new cv::cuda::GpuMat(params.m_imageHeight, params.m_imageWidth, CV_8UC4);
	depth_tex_ = new ftl::cuda::TextureObject<float>((cv::cuda::PtrStepSz<float>)*depth_mat_);
	colour_tex_ = new ftl::cuda::TextureObject<uchar4>((cv::cuda::PtrStepSz<uchar4>)*colour_mat_);
	data.depth = depth_tex_->cudaTexture();
	data.colour = colour_tex_->cudaTexture();
	data.params = params;

	if (withNormals) {
		point_mat_ = new cv::cuda::GpuMat(params.m_imageHeight, params.m_imageWidth, CV_32FC3);
		normal_mat_ = new cv::cuda::GpuMat(params.m_imageHeight, params.m_imageWidth, CV_32FC4);
		normal_tex_ = new ftl::cuda::TextureObject<float4>((cv::cuda::PtrStepSz<float4>)*normal_mat_);
		data.normal = normal_tex_->cudaTexture();
	} else {
		data.normal = 0;
	}
}

void DepthCamera::free() {
	if (depth_mat_) delete depth_mat_;
	if (colour_mat_) delete colour_mat_;
	if (point_mat_) delete point_mat_;
	if (normal_mat_) delete normal_mat_;
	delete depth_tex_;
	delete colour_tex_;
	if (normal_tex_) delete normal_tex_;
}

void DepthCamera::updateData(const cv::Mat &depth, const cv::Mat &rgb, cv::cuda::Stream &stream) {
	depth_mat_->upload(depth, stream);
	colour_mat_->upload(rgb, stream);
	if (normal_mat_) {
		_computeNormals(cv::cuda::StreamAccessor::getStream(stream));
	}
}

void DepthCamera::_computeNormals(cudaStream_t stream) {
	ftl::cuda::point_cloud((float3*)point_mat_->data, data, stream);
	ftl::cuda::compute_normals((float3*)point_mat_->data, normal_tex_, stream);
}
